/*
 * IcmpEchoResponse.java 
 *
 * Copyright (c) Verax Systems.
 * All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 */
package com.veraxsystems.icmp.jni;

import com.veraxsystems.icmp.IcmpStatus;
import com.veraxsystems.icmp.util.TimeUtil;

import java.net.InetAddress;
import java.util.TreeMap;

/**
 * Class defines content of a data received as a response of ICMP EchoRequest. 
 * It contains only important attributes of ICMP message that are 
 * sufficient to recognize result of ICMP EchoRequest processing ({@link #getStatus}).
 */
public class IcmpEchoResponse
{
   private final static byte DatagramConversionError = 31;
   
   private int         messageId;
   private byte        type;
   private byte        code;
   private byte[]      data;
   private InetAddress sourceAddress;
   private InetAddress destinationAddress;
   private long        creationTime;

   
   /**
    * Constructor for ICMP response. The creation time is automatically assigned 
    * during object construction. It is assumed, that response creation time is equal
    * to response receive time
    */
   public IcmpEchoResponse()
   {
      creationTime = TimeUtil.currentTimeMillis();
   }

   /*
    * The map of ICMP status integer codes so integer value may be easily 
    * converted to {@link IcmpStatus} enum.
    */
   private static TreeMap<Integer, IcmpStatus> icmpStatusMap = null;
   
   static {
      icmpStatusMap = new TreeMap<Integer, IcmpStatus>();
      for (IcmpStatus status: IcmpStatus.values()) {
         icmpStatusMap.put(new Integer(status.getValue()), status);
      }
   }

   /*
    * Convert integer value to coresponsing {@link IcmpStatus} enum.
    */
   private static IcmpStatus icmpStatusByInt(int value) 
   {
      IcmpStatus result = icmpStatusMap.get(value);
      if (result == null) result = IcmpStatus.UnexpectedResponse;
      return result;
   }
 
   /*
    * Get ICMP message code (see RFC792)
    */
   private byte getCode() {
      return code;
   }
   
   /**
    * Get data received in ICMP response message. This value is available
    * only when <code>EchoReply</code> message is received.
    * @return data received
    */
   public byte[] getData() {
      return data;
   }
   
   
   /**
    * Get ICMP EchoRequest message id combined as integer from identifier+sequenceNumber (see RFC792).
    * Used as part of filter for unique message identification.
    * @return identifier of ICMP message
    */
   public int getMessageId() {
      return messageId;
   }
   
   
   /*
    * Get ICMP message type (see RFC792)
    */
   private byte getType() {
      return type;
   }
   
   /**
    * Get address of host that was a source of ICMP response. On successful EchoRequest
    * processing it is the address of host provided as destination in ICMP EchoRequest.
    * In case network rejected the request, it is address of device (host) that has
    * rejected the message.
    * @return network address
    */
   public InetAddress getSourceAddress() {
      return sourceAddress;
   }
      
   /**
    * Get address of host the ICMP EchoRequest was sent to.
    * @return network address
    */
   public InetAddress getDestinationAddress() {
      return destinationAddress;
   }
      
   /**
    * Get status of ICMP response. The result enumeration describes type of ICMP response and is calculated
    * from <i>type</i> and <i>code</i> attributes of ICMP response message as defined in RFC 792. 
    * @return status if ICMP processing 
    */
   public IcmpStatus getStatus()
   {
      IcmpStatus result;
      
      // Class reports ICMP type 31 (Datagram Conversion Error) without detailed error code so no code is added
      // to status value
      if (getType() != DatagramConversionError)
         result = icmpStatusByInt(getType() * 256 + getCode());
      else
         result = icmpStatusByInt(getType() * 256);
      return result;
   }


   /**
    * Get time response was received. It is assumed creation time is equal to response time.
    * @return time response was received. Value in milliseconds
    */
   public long getResponseTime() 
   {
      return creationTime;
   }
   
}
